@@ -53,7 +53,6 @@ extern ApplicationData appdata;
 extern Cursor	  appcursors[NUM_CURSORS];
 extern fontinfo *fonts;
 extern short fontcount;
-extern short textpos, textend;
 
 #ifdef TCL_WRAPPER
 extern Tcl_Interp *xcinterp;
@@ -188,12 +187,13 @@ void changetextscale(float newscale)
 
    if (eventmode == TEXT_MODE || eventmode == ETEXT_MODE) {
       settext = *((labelptr *)EDITPART);
-      if (textpos > 0 || textpos < stringlength(settext->string, True,
+      if (areawin->textpos > 0 || areawin->textpos < stringlength(settext->string, True,
 		areawin->topinstance)) {
 	 undrawtext(settext);
-	 strptr = findstringpart(textpos - 1, NULL, settext->string,
+	 strptr = findstringpart(areawin->textpos - 1, NULL, settext->string,
+			areawin->topinstance);
+	 nextptr = findstringpart(areawin->textpos, NULL, settext->string,
 			areawin->topinstance);
-	 nextptr = findstringpart(textpos, NULL, settext->string, areawin->topinstance);
 	 if (strptr->type == FONT_SCALE)
 	    strptr->data.scale = newscale;
 	 else if (nextptr && nextptr->type == FONT_SCALE)
@@ -211,6 +211,7 @@ void changetextscale(float newscale)
 
    else if (areawin->selects > 0) {
       float oldscale;
+      Boolean waschanged = FALSE;
       for (osel = areawin->selectlist; osel < areawin->selectlist +
 	     areawin->selects; osel++) {
 	 if (SELECTTYPE(osel) == LABEL) {
@@ -220,17 +221,18 @@ void changetextscale(float newscale)
 	       undrawtext(settext);
                settext->scale = newscale;
 	       redrawtext(settext);
-	       register_for_undo(XCF_Scale, UNDO_MORE, areawin->topinstance,
+	       register_for_undo(XCF_Rescale, UNDO_MORE, areawin->topinstance,
 			(genericptr)settext, (double)oldscale);
+	       waschanged = TRUE;
 	    }
 	 }
       }
+      if (waschanged) undo_finish_series();
    }
 }
 
 /*--------------------------------------------------------------*/
 /* Auto-scale the drawing to fit the declared page size.	*/
-/* (for now, with fixed 1" margins)				*/
 /*								*/
 /* If the page is declared encapsulated, then do nothing.	*/
 /* If a frame box is on the page, then scale to fit the frame	*/
@@ -242,7 +244,6 @@ void autoscale(int page)
    float newxscale, newyscale;
    float scalefudge = (xobjs.pagelist[page]->coordstyle
 	== CM) ? CMSCALE : INCHSCALE;
-   float margin = 144.0;
    int width, height;
    polyptr framebox;
 
@@ -252,10 +253,10 @@ void autoscale(int page)
       return;
    }
 
-   newxscale = (xobjs.pagelist[page]->pagesize.x - margin)
-	/ scalefudge;
-   newyscale = (xobjs.pagelist[page]->pagesize.y - margin)
-	/ scalefudge;
+   newxscale = (xobjs.pagelist[page]->pagesize.x -
+	(2 * xobjs.pagelist[page]->margins.x)) / scalefudge;
+   newyscale = (xobjs.pagelist[page]->pagesize.y -
+	(2 * xobjs.pagelist[page]->margins.y)) / scalefudge;
 
    if ((framebox = checkforbbox(topobject)) != NULL) {
       int i, minx, miny, maxx, maxy;
@@ -321,6 +322,7 @@ float parseunits(char *strptr)
 /* Set the output page size, in the current unit of measure	*/
 /* Return value:  TRUE if _STR2 values were in inches, FALSE	*/
 /* if in centimeters.						*/
+/* XXX This API gives no good way to signal errors.             */
 /*--------------------------------------------------------------*/
 
 Boolean setoutputpagesize(XPoint *dataptr)
@@ -337,14 +339,14 @@ Boolean setoutputpagesize(XPoint *dataptr)
       if (sscanf(_STR2, "%f %*c %f", &px, &py) < 3) {
 	 if ((expos = strchr(_STR2, 'x')) == NULL) {
             Wprintf("Illegal Form for page size.");
-	    return;
+	    return FALSE;
 	 }
 	 else {
 	    *expos = '\0';
 	    if (sscanf(_STR2, "%f", &px) == 0 ||
 		  sscanf(expos + 1, "%f %9s", &py, units) == 0) {
                Wprintf("Illegal Form for page size.");
-	       return;
+	       return FALSE;
 	    }
 	 }
       }
@@ -355,7 +357,7 @@ Boolean setoutputpagesize(XPoint *dataptr)
 
    if ((px <= 2.0) || (py <= 2.0)) {
       Wprintf("Page size too small for margins.");
-      return;
+      return FALSE;
    }
 
    dataptr->x = (short)(px * 72.0);
@@ -383,12 +385,12 @@ labelptr gettextsize(float **floatptr)
    if (floatptr) *floatptr = &areawin->textscale;
 
    if (eventmode == TEXT_MODE || eventmode == ETEXT_MODE) {
-      if (textpos > 0 || textpos < stringlength(settext->string, True,
-		areawin->topinstance)) {
+      if (areawin->textpos > 0 || areawin->textpos < stringlength(settext->string,
+		True, areawin->topinstance)) {
          settext = *((labelptr *)EDITPART);
-	 strptr = findstringpart(textpos - 1, NULL, settext->string,
+	 strptr = findstringpart(areawin->textpos - 1, NULL, settext->string,
 			areawin->topinstance);
-	 nextptr = findstringpart(textpos, NULL, settext->string,
+	 nextptr = findstringpart(areawin->textpos, NULL, settext->string,
 			areawin->topinstance);
 	 if (strptr->type == FONT_SCALE) {
 	    if (floatptr) *floatptr = &strptr->data.scale;
@@ -471,10 +473,16 @@ void setdscale(xcWidget w, XPoint *dataptr)
 void setosize(xcWidget w, objinstptr dataptr)
 {
    float tmpres, oldsize;
+   Boolean waschanged = FALSE;
    short *osel;
    objinstptr nsobj;
    int res = sscanf(_STR2, "%f", &tmpres);
 
+   // Negative values are flips---deal with them independently
+
+   if (tmpres < 0)
+      tmpres = -tmpres;
+
    if (res == 0 || tmpres <= 0) {
       Wprintf("Illegal value");
       return;
@@ -484,14 +492,17 @@ void setosize(xcWidget w, objinstptr dataptr)
       if (SELECTTYPE(osel) == OBJINST) {
 	 nsobj = SELTOOBJINST(osel);
 	 oldsize = nsobj->scale;
-         nsobj->scale = tmpres;
+         nsobj->scale = (oldsize < 0) ? -tmpres : tmpres;
 
-         if (oldsize != tmpres)
-            register_for_undo(XCF_Scale, UNDO_MORE, areawin->topinstance,
+         if (oldsize != tmpres) {
+            register_for_undo(XCF_Rescale, UNDO_MORE, areawin->topinstance,
 			SELTOGENERIC(osel), (double)oldsize);
+	    waschanged = TRUE;
+	 }
       }
    }
-   unselect_all();
+   /* unselect_all(); */
+   if (waschanged) undo_finish_series();
    pwriteback(areawin->topinstance);
    drawarea(NULL, NULL, NULL);
 }
@@ -542,7 +553,7 @@ void setwwidth(xcWidget w, void *dataptr)
          }
 
 	 if (oldwidth != tmpres)
-            register_for_undo(XCF_Scale, UNDO_MORE, areawin->topinstance,
+            register_for_undo(XCF_Rescale, UNDO_MORE, areawin->topinstance,
 			SELTOGENERIC(osel), (double)oldwidth);
       }
       unselect_all();
@@ -570,16 +581,6 @@ void makenewfontbutton()
    fontnumbers[nfontnumbers - 1] = fontcount - 1;
 }
 
-/*--------------------------------------------------------------*/
-/* Set the menu checkmarks on the font menu			*/
-/*--------------------------------------------------------------*/
-
-void togglefontmark(int fontval)
-{
-   Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),fontfamily",
-		fonts[fontval].family, TCL_NAMESPACE_ONLY);
-}
-
 #endif /* TCL_WRAPPER */
 
 /*---------------------------------------------------------------*/
@@ -620,6 +621,7 @@ void setcolorscheme(Boolean boolvalue)
    INFOLABELCOLOR = appdata.infocolor;
    RATSNESTCOLOR = appdata.ratsnestcolor;
    BBOXCOLOR = appdata.bboxpix;
+   CLIPMASKCOLOR = appdata.clipcolor;
 }
 
 /*----------------------------------------------------------------*/
@@ -634,15 +636,10 @@ void togglegrid(u_short type)
       "decimal inches",
       "fractional inches",
       "centimeters",
-      "internal units",
+      "internal units", NULL
    };
 
-   /* Note that Tcl_SetVar() doesn't require double-quotes around the	*/
-   /* string; doing so prevents the menu from recognizing the variable	*/
-   /* value.								*/
-
-   Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),gridstyle",
-	stylenames[type], TCL_NAMESPACE_ONLY);
+   XcInternalTagCall(xcinterp, 3, "config", "coordstyle", stylenames[type]);
 }
 
 #endif /* TCL_WRAPPER */
@@ -741,11 +738,11 @@ void newpagemenu(xcWidget w, pointertype value, caddr_t nulldata)
 
 int createlibrary(Boolean force)
 {
-   xcWidget libmenu, newbutton, oldbutton;
+  /* xcWidget libmenu, newbutton, oldbutton; (jdk) */
 #ifndef TCL_WRAPPER
    Arg wargs[2];
 #endif
-   char libstring[20];
+   /* char libstring[20]; (jdk) */
    int libnum;
    objectptr newlibobj;
 
@@ -771,8 +768,6 @@ int createlibrary(Boolean force)
    xobjs.userlibs[libnum - LIBRARY].library = (objectptr *) malloc(sizeof(objectptr));
    xobjs.userlibs[libnum - LIBRARY].number = 0;
    xobjs.userlibs[libnum - LIBRARY].instlist = NULL;
-   xobjs.userlibs[libnum - LIBRARY].filename = NULL;
-   xobjs.userlibs[libnum - LIBRARY].flags = (char)0;
 
    sprintf(_STR2, "xcircuit::newlibrarybutton \"%s\"", newlibobj->name);
    Tcl_Eval(xcinterp, _STR2);
@@ -791,8 +786,8 @@ int createlibrary(Boolean force)
 
 void makepagebutton()
 {
-   xcWidget pagemenu, newbutton;
-   char pagestring[10];
+  /* xcWidget pagemenu, newbutton; (jdk) */
+  /* char pagestring[10]; (jdk) */
 
    /* make new entry in the menu */
 
@@ -812,7 +807,6 @@ void makepagebutton()
 
 void renamepage(short pagenumber)
 {
-   int page;
    objinstptr thisinst = xobjs.pagelist[pagenumber]->pageinst;
    char *pname, *plabel;
 
@@ -824,36 +818,6 @@ void renamepage(short pagenumber)
       Tcl_Eval(xcinterp, pname);
       free(pname);
    }
-
-   /* The GUI has been taken care of, but the page itself might not exist */
-   if (!thisinst) return;
-
-   /* If the page had a master but now has a different	*/
-   /* name, disconnect the schematics.			*/
-
-   if (thisinst->thisobject->schemtype == SECONDARY) {
-      if (strcmp(thisinst->thisobject->name, thisinst->thisobject->symschem->name)) {
-	 thisinst->thisobject->symschem = NULL;
-	 thisinst->thisobject->schemtype = PRIMARY;
-      }
-      Wprintf("Page disconnected from master schematic.");
-   }
-
-   /* Conversely, if the page has the same name as another page, the	*/
-   /* schematics are considered connected.				*/
-
-   for (page = 0; page < xobjs.pages; page++) {
-      objinstptr pageinst;
-      if (page == pagenumber) continue;
-      pageinst = xobjs.pagelist[page]->pageinst;
-      if (pageinst && !strcmp(pageinst->thisobject->name,
-		thisinst->thisobject->name)) {
-	 thisinst->thisobject->symschem = pageinst->thisobject;
-	 thisinst->thisobject->schemtype = SECONDARY;
-         Wprintf("Page connected to master schematic.");
-	 break;
-      }
-   }
 }
 
 /*--------------------------------------------------------------*/
@@ -877,24 +841,22 @@ void setcolormark(int colorval)
 {
    /* Set GUI variables and execute any command tags associated */
    /* with the "color" command */
-
-   Tcl_Obj *objv[3];
+  
+   char cstr[6];
    int i;
 
-   objv[0] = Tcl_NewStringObj("color", 5);
-   objv[1] = Tcl_NewStringObj("set", 3);
-   if (colorval == DEFAULTCOLOR)
-      objv[2] = Tcl_NewStringObj("inherit", 7);
-   else {
+   if (colorval != DEFAULTCOLOR) {
       for (i = 0; i < number_colors; i++)
-         if (colorlist[i].color.pixel == colorval)
+         if (colorlist[i].color.pixel == colorval) {
+	    sprintf(cstr, "%5d", i);
 	    break;
-      objv[2] = Tcl_NewIntObj((int)i);
+	 }
    }
-   Tcl_SetVar2Ex(xcinterp, "XCWinOps", "$XCOps(focus),colorval", objv[2],
-		TCL_NAMESPACE_ONLY);
 
-   XcTagCallback(xcinterp, 3, objv);
+#ifdef TCL_WRAPPER
+   XcInternalTagCall(xcinterp, 3, "color", "set", (colorval == DEFAULTCOLOR) ?
+	"inherit" : cstr);
+#endif
 }
 
 /*----------------------------------------------------------------*/
@@ -903,41 +865,25 @@ void setcolormark(int colorval)
 
 void setallstylemarks(u_short styleval)
 {
-   /* Set GUI variables and execute any command tags associated	*/
-   /* with the "fill" and "border" commands.			*/
+   /* Execute any command tags associated	*/
+   /* with the "fill" and "border" commands.	*/
 
-   Tcl_Obj *objv[3];
+   char fstr[10];
    int fillfactor;
    const char *bptr;
 
-   const char *borders[] = {"solid", "unbordered", "dashed", "dotted"};
+   const char *borders[] = {"solid", "unbordered", "dashed", "dotted", NULL};
    enum BorderIdx { SolidIdx, UnborderedIdx, DashedIdx, DottedIdx };
 
-   objv[0] = Tcl_NewStringObj("fill", 4);
-
    if (styleval & FILLED) {
       fillfactor = (int)(12.5 * (float)(1 + ((styleval & FILLSOLID) >> 5)));
-      if (fillfactor == 100)
-         objv[1] = Tcl_NewStringObj("solid", 5);
+      if (fillfactor < 100)
+	 sprintf(fstr, "%d", fillfactor);
       else
-         objv[1] = Tcl_NewIntObj(fillfactor);
+	 strcpy(fstr, "solid");
    }
    else
-      objv[1] = Tcl_NewStringObj("unfilled", 8);
-
-   if (styleval & OPAQUE)
-      objv[2] = Tcl_NewStringObj("opaque", 6);
-   else
-      objv[2] = Tcl_NewStringObj("transparent", 11);
-
-   Tcl_SetVar2Ex(xcinterp, "XCWinOps", "$XCOps(focus),fillamount", objv[1],
-		TCL_NAMESPACE_ONLY);
-   Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),opaque", (styleval & OPAQUE) ?
-		"true" : "false", TCL_NAMESPACE_ONLY);
-   Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),bboxtype", (styleval & BBOX) ?
-		"true" : "false", TCL_NAMESPACE_ONLY);
-   Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),polyclosed",
-		(styleval & UNCLOSED) ? "false" : "true", TCL_NAMESPACE_ONLY);
+      strcpy(fstr, "unfilled");
 
    switch (styleval & BORDERS) {
       case DASHED:
@@ -953,10 +899,16 @@ void setallstylemarks(u_short styleval)
 	 bptr = borders[SolidIdx];
 	 break;
    }
-   Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),linestyle", (char *)bptr,
-		TCL_NAMESPACE_ONLY);
 
-   XcTagCallback(xcinterp, 3, objv);
+   XcInternalTagCall(xcinterp, 3, "fill", fstr,
+		(styleval & OPAQUE) ? "opaque" : "transparent");
+   XcInternalTagCall(xcinterp, 3, "border", "bbox", (styleval & BBOX) ? "true" :
+		"false");
+   XcInternalTagCall(xcinterp, 3, "border", "clipmask", (styleval & CLIPMASK) ?
+		"true" : "false");
+   XcInternalTagCall(xcinterp, 2, "border", (styleval & UNCLOSED) ? "unclosed" :
+		"closed");
+   XcInternalTagCall(xcinterp, 2, "border", bptr);
 }
 
 #endif /* TCL_WRAPPER */
@@ -985,10 +937,19 @@ polyptr checkforbbox(objectptr localdata)
 
 int setelementstyle(xcWidget w, u_short value, u_short mask)
 {
-   Boolean selected = False;
+   Boolean preselected, selected = False;
    short *sstyle;
    u_short newstyle, oldstyle;
 
+   if (areawin->selects == 0) {
+      preselected = FALSE;
+      if (value & BBOX)
+	 checkselect(POLYGON);
+      else
+	 checkselect(ARC | SPLINE | POLYGON | PATH);
+   }
+   else preselected = TRUE;
+
    if (areawin->selects > 0) {
       if (value & BBOX) {
 	 polyptr ckp;
@@ -1050,7 +1011,9 @@ int setelementstyle(xcWidget w, u_short value, u_short mask)
 	       easydraw(*sstyle, DOFORALL);
 
 #ifdef TCL_WRAPPER
-	       register_for_undo(XCF_ChangeStyle, UNDO_DONE, areawin->topinstance,
+	       register_for_undo(XCF_ChangeStyle,
+			(sstyle == areawin->selectlist + areawin->selects - 1) ?
+			UNDO_DONE : UNDO_MORE, areawin->topinstance,
 			SELTOGENERIC(sstyle), (int)oldstyle);
 #else
 	       /* Tcl version differs in that the element is not deselected after */
@@ -1072,6 +1035,11 @@ int setelementstyle(xcWidget w, u_short value, u_short mask)
 	 newstyle &= ~(BBOX);
 	 return -1;
       }
+      else if (value & CLIPMASK) {
+	 Wprintf("Cannot set default style to Clip Mask");
+	 newstyle &= ~(CLIPMASK);
+	 return -1;
+      }
       else {
 	 newstyle &= ~mask;
 	 newstyle |= value;
@@ -1088,8 +1056,10 @@ int setelementstyle(xcWidget w, u_short value, u_short mask)
    }
 #ifndef TCL_WRAPPER
    setallstylemarks(newstyle);
-   unselect_all();
 #endif
+   if (!preselected)
+      unselect_all();
+
    return (int)newstyle;
 }
 
@@ -1102,7 +1072,7 @@ int setelementstyle(xcWidget w, u_short value, u_short mask)
 void setcolor(xcWidget w, int cindex)
 {
    short *scolor;
-   int *ecolor, cval;
+   int *ecolor, cval, oldcolor;
    Boolean selected = False;
    stringpart *strptr, *nextptr;
 
@@ -1111,8 +1081,10 @@ void setcolor(xcWidget w, int cindex)
 
    if (eventmode == TEXT_MODE || eventmode == ETEXT_MODE) {
       labelptr curlabel = TOLABEL(EDITPART);
-      strptr = findstringpart(textpos - 1, NULL, curlabel->string, areawin->topinstance);
-      nextptr = findstringpart(textpos, NULL, curlabel->string, areawin->topinstance);
+      strptr = findstringpart(areawin->textpos - 1, NULL, curlabel->string,
+			areawin->topinstance);
+      nextptr = findstringpart(areawin->textpos, NULL, curlabel->string,
+			areawin->topinstance);
       if (strptr->type == FONT_COLOR) {
 	 undrawtext(curlabel);
 	 strptr->data.color = cindex;
@@ -1133,9 +1105,14 @@ void setcolor(xcWidget w, int cindex)
       for (scolor = areawin->selectlist; scolor < areawin->selectlist
 	   + areawin->selects; scolor++) {
 	 ecolor = &(SELTOCOLOR(scolor));
-
+	 oldcolor = *ecolor;
 	 *ecolor = cval;
 	 selected = True;
+
+	 register_for_undo(XCF_Color, (scolor == areawin->selectlist
+			+ areawin->selects - 1) ?  UNDO_DONE : UNDO_MORE,
+			areawin->topinstance,
+			SELTOGENERIC(scolor), (int)oldcolor);
       }
    }
 
@@ -1144,6 +1121,17 @@ void setcolor(xcWidget w, int cindex)
       if (eventmode != TEXT_MODE && eventmode != ETEXT_MODE)
          areawin->color = cval;      
    }
+   else pwriteback(areawin->topinstance);
+}
+
+/*--------------------------------------------------------------*/
+/* Set the menu checkmarks on the font menu			*/
+/*--------------------------------------------------------------*/
+
+void togglefontmark(int fontval)
+{
+   if (fonts[fontval].family != NULL)
+      XcInternalTagCall(xcinterp, 3, "label", "family", fonts[fontval].family);
 }
 
 /*------------------------------------------------------*/
@@ -1153,106 +1141,51 @@ void setcolor(xcWidget w, int cindex)
 
 void togglestylemark(int styleval)
 {
-   switch(styleval) {
-      case 0: Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),fontstyle",
-		"normal", TCL_NAMESPACE_ONLY);
-	 break;
-      case 1: Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),fontstyle",
-		"bold", TCL_NAMESPACE_ONLY);
-	 break;
-      case 2: Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),fontstyle",
-		"italic", TCL_NAMESPACE_ONLY);
-	 break;
-      case 3: Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),fontstyle",
-		"bolditalic", TCL_NAMESPACE_ONLY);
-	 break;
-   }
+   char *cstyle = translatestyle(styleval);
+   if (cstyle != NULL)
+      XcInternalTagCall(xcinterp, 3, "label", "style", cstyle);
 }
 
+/*------------------------------------------------------*/
+/* Set checkmarks on label encoding menu		*/
 /*------------------------------------------------------*/
 
 void toggleencodingmark(int encodingval)
 {
-   switch(encodingval) {
-      case 0: Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),fontencoding",
-		"Standard", TCL_NAMESPACE_ONLY);
-	 break;
-      case 1: Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),fontencoding",
-		"special", TCL_NAMESPACE_ONLY);
-	 break;
-      case 2: Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),fontencoding",
-		"ISOLatin1", TCL_NAMESPACE_ONLY);
-	 break;
-      case 3: Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),fontencoding",
-		"ISOLatin2", TCL_NAMESPACE_ONLY);
-	 break;
-      case 4: Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),fontencoding",
-		"ISOLatin3", TCL_NAMESPACE_ONLY);
-	 break;
-      case 5: Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),fontencoding",
-		"ISOLatin4", TCL_NAMESPACE_ONLY);
-	 break;
-      case 6: Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),fontencoding",
-		"ISOLatin5", TCL_NAMESPACE_ONLY);
-	 break;
-      case 7: Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),fontencoding",
-		"ISOLatin6", TCL_NAMESPACE_ONLY);
-	 break;
-   }
+   char *cenc = translateencoding(encodingval);
+   if (cenc != NULL)
+      XcInternalTagCall(xcinterp, 3, "label", "encoding", cenc);
 }
 
+/*------------------------------------------------------*/
+/* Set checkmarks on label justification & flags menu	*/
 /*------------------------------------------------------*/
 
 void togglejustmarks(int justvalue)
 {
-   switch(justvalue & RLJUSTFIELD) {
-      case NORMAL: Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),jhoriz",
-		"left", TCL_NAMESPACE_ONLY);
-	 break;
-      case NOTLEFT: Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),jhoriz",
-		"center", TCL_NAMESPACE_ONLY);
-	 break;
-      case RIGHT|NOTLEFT: Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),jhoriz",
-		"right", TCL_NAMESPACE_ONLY);
-	 break;
-   }
-   switch(justvalue & TBJUSTFIELD) {
-      case NORMAL: Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),jvert",
-		"bottom", TCL_NAMESPACE_ONLY);
-	 break;
-      case NOTBOTTOM: Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),jvert",
-		"middle", TCL_NAMESPACE_ONLY);
-	 break;
-      case TOP|NOTBOTTOM: Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),jvert",
-		"top", TCL_NAMESPACE_ONLY);
-	 break;
-   }
-
-   /* Flip Invariance property */
-   Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),flipinvariant",
-		(justvalue & FLIPINV) ?  "true" : "false", TCL_NAMESPACE_ONLY);
-
-   /* LaTeX mode property */
-   Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),latexmode",
-		(justvalue & LATEXLABEL) ?  "true" : "false", TCL_NAMESPACE_ONLY);
-     
-   /* Pin visibility property */
-   Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),pinvisible",
-		(justvalue & PINVISIBLE) ?  "true" : "false", TCL_NAMESPACE_ONLY);
+   XcInternalTagCall(xcinterp, 4, "label", "justify", (justvalue & RIGHT) ? "right" :
+	(justvalue & NOTLEFT) ? "center" : "left",
+	(justvalue & TOP) ? "top" : (justvalue & NOTBOTTOM) ? "middle" : "bottom");
+   XcInternalTagCall(xcinterp, 3, "label", "flipinvariant", (justvalue & FLIPINV) ?
+	"true" : "false");
+   XcInternalTagCall(xcinterp, 3, "label", "latex", (justvalue & LATEXLABEL) ?
+	"true" : "false");
+   XcInternalTagCall(xcinterp, 3, "label", "visible", (justvalue & PINVISIBLE) ?
+	"true" : "false");
 }
 
-/*------------------------------------------------------*/
+/*-------------------------------------------------------------*/
+/* Simultaneously set all relevant checkmarks for a text label */
+/*-------------------------------------------------------------*/
 
 void setfontmarks(short fvalue, short jvalue)
 {
-   if ((fvalue >= 0) && (fontcount > 0)) {
-      togglestylemark(fonts[fvalue].flags & 0x03);
-      toggleencodingmark((fonts[fvalue].flags & 0xf80) >> 7);
+   if ((fvalue >= 0) && (fvalue < fontcount)) {
+      toggleencodingmark(fvalue);
+      togglestylemark(fvalue);
       togglefontmark(fvalue);
    }
-   if (jvalue >= 0) {
-      togglejustmarks(jvalue);
-   }
+   togglejustmarks(jvalue);
 }
 
 #endif /* TCL_WRAPPER */
@@ -1267,13 +1200,13 @@ void stringparam(xcWidget w, caddr_t clientdata, caddr_t calldata)
 {
    genericptr *settext;
 
-      if (eventmode == TEXT_MODE || eventmode == ETEXT_MODE) {
-         settext = (genericptr *)EDITPART;
-         makeparam(TOLABEL(settext), _STR2);
-         unselect_all();
-	 setparammarks(NULL);
-      }
-      else if (checkselect(LABEL)) parameterize(P_SUBSTRING, _STR2);
+   if (eventmode == TEXT_MODE || eventmode == ETEXT_MODE) {
+      settext = (genericptr *)EDITPART;
+      makeparam(TOLABEL(settext), _STR2);
+      unselect_all();
+      setparammarks(NULL);
+   }
+   else if (checkselect(LABEL)) parameterize(P_SUBSTRING, _STR2, -1);
 }
 
 /*--------------------------------------------------------------*/
@@ -1282,14 +1215,12 @@ void stringparam(xcWidget w, caddr_t clientdata, caddr_t calldata)
 
 void startparam(xcWidget w, pointertype value, caddr_t calldata)
 {
-   genericptr *settext;
-
    if (value == (pointertype)P_SUBSTRING) {
       strcpy(_STR2, (calldata != NULL) ? (char *)calldata : "substring");
       stringparam(w, NULL, NULL);
    }
    else if ((eventmode != NORMAL_MODE) || (areawin->selects > 0))
-	 parameterize((int)value, (char *)calldata);
+	 parameterize((int)value, (char *)calldata, -1);
 }
 
 /*---------------------------------------------------------------*/
@@ -1313,261 +1244,6 @@ void setdefaultfontmarks()
    setfontmarks(areawin->psfont, areawin->justify);
 }
 
-/*--------------------------------------------------------------*/
-/* Toggle a bit in the justification field of a label		*/
-/*--------------------------------------------------------------*/
-
-void dojustifybit(xcWidget w, labelptr settext, short bitfield)
-{
-   if (settext != NULL) {
-      int oldjust = (int)settext->justify;
-
-      undrawtext(settext);
-      settext->justify ^= bitfield;
-      redrawtext(settext);
-      pwriteback(areawin->topinstance);
-
-      register_for_undo(XCF_Justify, UNDO_MORE, areawin->topinstance,
-		(genericptr)settext, oldjust);
-   }
-   else
-      areawin->justify ^= bitfield;
-
-#ifndef TCL_WRAPPER
-   if (w != NULL) {
-      Boolean boolval;
-      if (settext)
-	 boolval = (settext->justify & bitfield) ? 0 : 1;
-      else
-	 boolval = (areawin->justify & bitfield) ? 0 : 1;
-      toggle(w, (pointertype)(-1), &boolval);
-   }
-#endif
-}
-
-/*--------------------------------------------------------------*/
-/* Toggle a pin-related bit in the justification field of a	*/
-/* label.  This differs from dojustifybit() in that the label	*/
-/* must be a pin, and this function cannot change the default	*/
-/* behavior set by areawin->justify.				*/
-/*--------------------------------------------------------------*/
-
-void dopinjustbit(xcWidget w, labelptr settext, short bitfield)
-{
-   if ((settext != NULL) && settext->pin) {
-      int oldjust = (int)settext->justify;
-
-      undrawtext(settext);
-      settext->justify ^= bitfield;
-      redrawtext(settext);
-      pwriteback(areawin->topinstance);
-
-      register_for_undo(XCF_Justify, UNDO_MORE, areawin->topinstance,
-		(genericptr)settext, oldjust);
-
-#ifndef TCL_WRAPPER
-      if (w != NULL) {
-	 Boolean boolval = (settext->justify & bitfield) ? 0 : 1;
-	 toggle(w, (pointertype)(-1), &boolval);
-      }
-#endif
-   }
-}
-
-/*----------------------------------------------------------------*/
-/* Set the justification for the label passed as 3rd parameter	  */
-/*----------------------------------------------------------------*/
-
-void setjust(xcWidget w, pointertype value, labelptr settext, short mode)
-{
-   short newjust, oldjust;
-
-   if (settext != NULL) {
-
-      if (mode == 1)
-         newjust = (settext->justify & (NONJUSTFIELD | TBJUSTFIELD))
-			| value;
-      else
-         newjust = (settext->justify & (NONJUSTFIELD | RLJUSTFIELD))
-			| value;
-
-      if (settext->justify != newjust) {
-	  oldjust = (int)settext->justify;
-          undrawtext(settext);
-          settext->justify = newjust;
-          redrawtext(settext);
-          pwriteback(areawin->topinstance);
-
-          register_for_undo(XCF_Justify, UNDO_MORE, areawin->topinstance,
-		(genericptr)settext, oldjust);
-      }
-   }
-   else {
-      if (mode == 1)
-         newjust = (areawin->justify & (NONJUSTFIELD | TBJUSTFIELD))
-			| value;
-      else
-         newjust = (areawin->justify & (NONJUSTFIELD | RLJUSTFIELD))
-			| value;
-      areawin->justify = newjust;
-   }
-#ifndef TCL_WRAPPER
-   if (w != NULL) togglejustifs(w);
-#endif
-}
-
-/*----------------------------------------------------------------*/
-/* Set vertical justification (top, middle, bottom) on all	  */
-/* selected labels						  */
-/*----------------------------------------------------------------*/
-
-void setvjust(xcWidget w, pointertype value, caddr_t nulldata)
-{
-   short *fselect;
-   labelptr settext;
-   short labelcount = 0;
-
-   if (eventmode == TEXT_MODE || eventmode == ETEXT_MODE) {
-      settext = *((labelptr *)EDITPART);
-      setjust(w, value, settext, 2);
-   }
-   else {
-      for (fselect = areawin->selectlist; fselect < areawin->selectlist +
-            areawin->selects; fselect++) {
-         if (SELECTTYPE(fselect) == LABEL) {
-            labelcount++;
-            settext = SELTOLABEL(fselect);
-            setjust(NULL, value, settext, 2);
-         }
-      }
-      if (labelcount == 0) setjust(w, value, NULL, 2);
-      else unselect_all();
-   }
-}
-
-/*----------------------------------------------------------------*/
-/* Set horizontal justification (left, center, right) on all	  */
-/* selected labels						  */
-/*----------------------------------------------------------------*/
-
-void sethjust(xcWidget w, pointertype value, caddr_t nulldata)
-{
-   short *fselect;
-   labelptr settext;
-   short labelcount = 0;
-
-   if (eventmode == TEXT_MODE || eventmode == ETEXT_MODE) {
-      settext = *((labelptr *)EDITPART);
-      setjust(w, value, settext, 1);
-   }
-   else {
-      for (fselect = areawin->selectlist; fselect < areawin->selectlist +
-            areawin->selects; fselect++) {
-         if (SELECTTYPE(fselect) == LABEL) {
-            labelcount++;
-            settext = SELTOLABEL(fselect);
-            setjust(NULL, value, settext, 1);
-         }
-      }
-      if (labelcount == 0) setjust(w, value, NULL, 1);
-      else unselect_all();
-   }
-}
-
-/*--------------------------------------------------------------*/
-/* Set a justify field bit on all selected labels		*/ 
-/* (flip invariance, latex mode, etc.)				*/
-/*--------------------------------------------------------------*/
-
-void setjustbit(xcWidget w, pointertype value, caddr_t nulldata)
-{
-   short *fselect;
-   labelptr settext;
-   short labelcount = 0;
-
-   if (eventmode == TEXT_MODE || eventmode == ETEXT_MODE) {
-      settext = *((labelptr *)EDITPART);
-      dojustifybit(w, settext, (short)value);
-   }
-   else {
-      for (fselect = areawin->selectlist; fselect < areawin->selectlist +
-            areawin->selects; fselect++) {
-         if (SELECTTYPE(fselect) == LABEL) {
-            labelcount++;
-            settext = SELTOLABEL(fselect);
-            dojustifybit(NULL, settext, (short)value);
-         }
-      }
-      if (labelcount == 0) dojustifybit(w, NULL, (short)value);
-      else unselect_all();
-   }
-}
-
-/*--------------------------------------------------------------*/
-/* Set pin-related bit field of "justify" on all selected pins	*/ 
-/* (e.g., pin visibility)					*/
-/*--------------------------------------------------------------*/
-
-void setpinjustbit(xcWidget w, pointertype value, caddr_t nulldata)
-{
-   short *fselect;
-   labelptr settext;
-
-   if (eventmode == TEXT_MODE || eventmode == ETEXT_MODE) {
-      settext = *((labelptr *)EDITPART);
-      if (settext->pin)
-	 dopinjustbit(w, settext, (short)value);
-   }
-   else {
-      for (fselect = areawin->selectlist; fselect < areawin->selectlist +
-            areawin->selects; fselect++) {
-         if (SELECTTYPE(fselect) == LABEL) {
-            settext = SELTOLABEL(fselect);
-	    if (settext->pin)
-               dopinjustbit(NULL, settext, (short)value);
-         }
-      }
-      unselect_all();
-   }
-}
-
-/*---------------------------*/
-/* Set polygon editing style */
-/*---------------------------*/
-
-#ifdef TCL_WRAPPER
-
-void boxedit(xcWidget w, pointertype value, caddr_t nulldata)
-{
-   switch (value) {
-      case MANHATTAN:
-	 Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),polyedittype",
-		"manhattan", TCL_NAMESPACE_ONLY);
-	 break;
-      case RHOMBOIDX:
-	 Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),polyedittype",
-		"rhomboidx", TCL_NAMESPACE_ONLY);
-	 break;
-      case RHOMBOIDY:
-	 Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),polyedittype",
-		"rhomboidy", TCL_NAMESPACE_ONLY);
-	 break;
-      case RHOMBOIDA:
-	 Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),polyedittype",
-		"rhomboida", TCL_NAMESPACE_ONLY);
-	 break;
-      case NORMAL:
-	 Tcl_SetVar2(xcinterp, "XCWinOps", "$XCOps(focus),polyedittype",
-		"normal", TCL_NAMESPACE_ONLY);
-	 break;
-   }
-
-   if (areawin->boxedit == value) return;
-   areawin->boxedit = value;
-}
-
-#endif /* TCL_WRAPPER */
-
 /*----------------------------------------------------------------*/
 /* Pick up font name from _STR2 and pass it to loadfontfile()	  */
 /*----------------------------------------------------------------*/
@@ -1693,9 +1369,9 @@ void setfontval(xcWidget w, pointertype value, labelptr settext)
 
       /* if last byte was a font designator, use it */
 
-      if (textpos > 0 || textpos < stringlength(settext->string, True,
-			areawin->topinstance)) {
-	 strptr = findstringpart(textpos - 1, NULL, settext->string,
+      if (areawin->textpos > 0 || areawin->textpos < stringlength(settext->string,
+			True, areawin->topinstance)) {
+	 strptr = findstringpart(areawin->textpos - 1, NULL, settext->string,
 			areawin->topinstance);
 	 if (strptr->type == FONT_NAME) {
 	    tc = strptr->data.font;
@@ -1714,7 +1390,7 @@ void setfontval(xcWidget w, pointertype value, labelptr settext)
       }
 
       /* otherwise, look for the last style used in the string */
-      tc = findcurfont(textpos, settext->string, areawin->topinstance);
+      tc = findcurfont(areawin->textpos, settext->string, areawin->topinstance);
    }
    else tc = areawin->psfont;
 
@@ -1745,6 +1421,7 @@ void setfont(xcWidget w, pointertype value, caddr_t calldata)
    short *fselect;
    labelptr settext;
    short labelcount = 0;
+   Boolean preselected;
 
    if (eventmode == CATALOG_MODE || eventmode == FONTCAT_MODE ||
 		eventmode == EFONTCAT_MODE) return;
@@ -1755,7 +1432,12 @@ void setfont(xcWidget w, pointertype value, caddr_t calldata)
       charreport(settext);
    }
    else {
-      textpos = 1;
+      if (areawin->selects == 0) {
+	 checkselect(LABEL);
+	 preselected = FALSE;
+      }
+      else preselected = TRUE;
+      areawin->textpos = 1;
       for (fselect = areawin->selectlist; fselect < areawin->selectlist +
 	    areawin->selects; fselect++) {
          if (SELECTTYPE(fselect) == LABEL) {
@@ -1765,7 +1447,7 @@ void setfont(xcWidget w, pointertype value, caddr_t calldata)
          }
       }
       if (labelcount == 0) setfontval(w, value, NULL);
-      else unselect_all();
+      else if (!preselected) unselect_all();
    }
 }
 
@@ -1784,8 +1466,10 @@ void setfontstyle(xcWidget w, pointertype value, labelptr settext)
 
       /* if last byte was a font designator, use it */
 
-      if (textpos > 0 || textpos < stringlength(settext->string, True, areawin->topinstance)) {
-	 strptr = findstringpart(textpos - 1, NULL, settext->string, areawin->topinstance);
+      if (areawin->textpos > 0 || areawin->textpos < stringlength(settext->string,
+			True, areawin->topinstance)) {
+	 strptr = findstringpart(areawin->textpos - 1, NULL, settext->string,
+			areawin->topinstance);
 	 if (strptr->type == FONT_NAME) {
 	    tc = strptr->data.font;
 
@@ -1807,8 +1491,11 @@ void setfontstyle(xcWidget w, pointertype value, labelptr settext)
          }
       }
 
-      /* otherwise, look for the last font used in the string */
-      tc = findcurfont(textpos - 2, settext->string, areawin->topinstance);
+      /* Otherwise, look for the last font used in the string */
+      /* Fix by Dimitri Princen.  Was areawin->textpos - 2; was there a */
+      /* reason for that?						*/
+
+      tc = findcurfont(areawin->textpos, settext->string, areawin->topinstance);
    }
    else tc = areawin->psfont;
 
@@ -1839,6 +1526,7 @@ void fontstyle(xcWidget w, pointertype value, caddr_t nulldata)
    short *fselect;
    labelptr settext;
    short labelcount = 0;
+   Boolean preselected;
 
    if (eventmode == CATALOG_MODE || eventmode == FONTCAT_MODE ||
 		eventmode == EFONTCAT_MODE) return;
@@ -1849,7 +1537,12 @@ void fontstyle(xcWidget w, pointertype value, caddr_t nulldata)
       charreport(settext);
    }
    else {
-      textpos = 1;
+      if (areawin->selects == 0) {
+	 checkselect(LABEL);
+	 preselected = FALSE;
+      }
+      else preselected = TRUE;
+      areawin->textpos = 1;
       for (fselect = areawin->selectlist; fselect < areawin->selectlist +
 	    areawin->selects; fselect++) {
          if (SELECTTYPE(fselect) == LABEL) {
@@ -1859,7 +1552,7 @@ void fontstyle(xcWidget w, pointertype value, caddr_t nulldata)
          }
       }
       if (labelcount == 0) setfontstyle(w, value, NULL);
-      else unselect_all();
+      else if (!preselected) unselect_all();
    }
 }
 
@@ -1878,9 +1571,9 @@ void setfontencoding(xcWidget w, pointertype value, labelptr settext)
 
       /* if last byte was a font designator, use it */
 
-      if (textpos > 0 || textpos < stringlength(settext->string, True,
-			areawin->topinstance)) {
-	 strptr = findstringpart(textpos - 1, NULL, settext->string,
+      if (areawin->textpos > 0 || areawin->textpos < stringlength(settext->string,
+		True, areawin->topinstance)) {
+	 strptr = findstringpart(areawin->textpos - 1, NULL, settext->string,
 			areawin->topinstance);
 	 if (strptr->type == FONT_NAME) {
 	    tc = strptr->data.font;
@@ -1904,7 +1597,7 @@ void setfontencoding(xcWidget w, pointertype value, labelptr settext)
       }
 
       /* otherwise, look for the last style used in the string */
-      tc = findcurfont(textpos - 2, settext->string, areawin->topinstance);
+      tc = findcurfont(areawin->textpos - 2, settext->string, areawin->topinstance);
    }
    else tc = areawin->psfont;
 
@@ -1934,6 +1627,7 @@ void fontencoding(xcWidget w, pointertype value, caddr_t nulldata)
    short *fselect;
    labelptr settext;
    short labelcount = 0;
+   Boolean preselected;
 
    if (eventmode == CATALOG_MODE || eventmode == FONTCAT_MODE ||
 		eventmode == EFONTCAT_MODE) return;
@@ -1944,7 +1638,12 @@ void fontencoding(xcWidget w, pointertype value, caddr_t nulldata)
       charreport(settext);
    }
    else {
-      textpos = 1;
+      if (areawin->selects == 0) {
+  	 checkselect(LABEL);
+	 preselected = FALSE; 
+      }
+      else preselected = TRUE;
+      areawin->textpos = 1;
       for (fselect = areawin->selectlist; fselect < areawin->selectlist +
 	    areawin->selects; fselect++) {
          if (SELECTTYPE(fselect) == LABEL) {
@@ -1954,7 +1653,7 @@ void fontencoding(xcWidget w, pointertype value, caddr_t nulldata)
          }
       }
       if (labelcount == 0) setfontencoding(w, value, NULL);
-      else unselect_all();
+      else if (!preselected) unselect_all();
    }
 }
 
@@ -1975,7 +1674,7 @@ Boolean dospecial()
    curlabel = TOLABEL(EDITPART);
    if (curlabel->justify & LATEXLABEL) return False;
 
-   cfont = findcurfont(textpos, curlabel->string, areawin->topinstance);
+   cfont = findcurfont(areawin->textpos, curlabel->string, areawin->topinstance);
    composefontlib(cfont);
    startcatalog(NULL, FONTLIB, NULL);
    return True;
