@@ -7,8 +7,10 @@
 #include <stdlib.h>
 #include <ctype.h>   /* for isprint() and isdigit() */
 
+#ifndef _MSC_VER
 #include <X11/Intrinsic.h>
 #include <X11/StringDefs.h>
+#endif
 
 /*------------------------------------------------------------------------*/
 /* Local includes                                                         */
@@ -33,7 +35,6 @@
 extern Display *dpy;
 extern XCWindowData *areawin;
 extern Globaldata xobjs;
-extern short textpos, textend;
 extern short fontcount;
 extern fontinfo *fonts;
 extern int *appcolors;
@@ -54,7 +55,7 @@ short del;
 /* calling routine to free it.						*/
 /*----------------------------------------------------------------------*/
 
-char *evaluate_expr(oparamptr ops, objinstptr pinst)
+char *evaluate_expr(objectptr thisobj, oparamptr ops, objinstptr pinst)
 {
    if (ops->type != XC_EXPR) return NULL;
    return strdup(ops->parameter.expr);
@@ -83,10 +84,10 @@ Boolean hasparameter(labelptr curlabel)
 
 void joinlabels()
 {
-   genericptr *genobj;
+  /* genericptr *genobj; (jdk) */
    short *jl;
    stringpart *endpart;
-   int tpos;
+   /* int tpos; (jdk) */
    labelptr dest, source;
 
    if (areawin->selects < 2) {
@@ -151,10 +152,10 @@ stringpart *makesegment(stringpart **strhead, stringpart *before)
 	       char *key = lastptr->data.string;
 	       obs = find_param(areawin->topinstance, key);
 	       if (obs == NULL) {
-		  Wprintf("Error:  Bad parameter!");
-		  return NULL;
+		  Wprintf("Error:  Bad parameter \"%s\"!", key);
+	       } else {
+	          obs->parameter.string = newptr;	/* ?? */
 	       }
-	       obs->parameter.string = newptr;
 	    }
 	    else {
 	       lastptr->nextpart = newptr;
@@ -198,10 +199,10 @@ stringpart *splitstring(int tpos, stringpart **strtop, objinstptr localinst)
 }
 
 /*----------------------------------------------------------------------*/
-/* Get the next string part, linking to a paramter if necessary		*/
+/* Get the next string part, linking to a parameter if necessary	*/
 /*----------------------------------------------------------------------*/
 
-stringpart *nextstringpart(stringpart *strptr, objinstptr thisinst)
+stringpart *nextstringpartrecompute(stringpart *strptr, objinstptr thisinst)
 {
    stringpart *nextptr = strptr->nextpart;
 
@@ -229,7 +230,7 @@ stringpart *nextstringpart(stringpart *strptr, objinstptr thisinst)
 /* when encountered.  Use the previously generated result.		*/
 /*----------------------------------------------------------------------*/
 
-stringpart *nextstringpartnorecompute(stringpart *strptr, objinstptr thisinst)
+stringpart *nextstringpart(stringpart *strptr, objinstptr thisinst)
 {
    stringpart *nextptr = strptr->nextpart;
 
@@ -302,7 +303,8 @@ stringpart *deletestring(stringpart *dstr, stringpart **strtop, objinstptr thisi
       else
 	 strptr->nextpart = dstr->nextpart;
    }
-   if (dstr->type == TEXT_STRING) free(dstr->data.string);
+   if (dstr->type == TEXT_STRING)
+      free(dstr->data.string);
    free(dstr);
 
    /* attempt to merge, if legal */
@@ -345,6 +347,11 @@ stringpart *mergestring(stringpart *firststr)
 /* parameters encountered.  If FALSE, then we assume that all		*/
 /* expressions have been computed previously, and may use the recorded	*/
 /* instance value.							*/
+/*									*/
+/* 11/20/06---changed to allow two different static strings to save	*/
+/* promoted results.  This is necessary because we may be comparing	*/
+/* two promoted results in, e.g., stringcomprelaxed(), and we don't	*/
+/* want to overwrite the first result with the second.			*/
 /*----------------------------------------------------------------------*/
 
 stringpart *linkstring(objinstptr localinst, stringpart *strstart,
@@ -352,7 +359,8 @@ stringpart *linkstring(objinstptr localinst, stringpart *strstart,
 {
    char *key;
    stringpart *tmpptr, *nextptr = NULL;
-   static stringpart *promote = NULL;
+   static stringpart *promote[2] = {NULL, NULL};
+   static unsigned char pidx = 0;
    oparamptr ops;
 
    if (strstart->type != PARAM_START) return NULL;
@@ -364,7 +372,8 @@ stringpart *linkstring(objinstptr localinst, stringpart *strstart,
 
    if (localinst == NULL) {
       ops = match_param(topobject, key);
-      if (ops == NULL) return NULL;
+      if (ops == NULL)
+	 return NULL;
    }
    else {
       ops = find_param(localinst, key);
@@ -378,41 +387,48 @@ stringpart *linkstring(objinstptr localinst, stringpart *strstart,
 
    if (ops->type != XC_STRING) {
 
-      if (promote == NULL) {
+      if (promote[pidx] == NULL) {
          /* Generate static string for promoting numerical parameters */
-         tmpptr = makesegment(&promote, NULL);
+         tmpptr = makesegment(&promote[pidx], NULL);
          tmpptr->type = TEXT_STRING;
-         tmpptr = makesegment(&promote, NULL);
+         tmpptr = makesegment(&promote[pidx], NULL);
          tmpptr->type = PARAM_END;
       }
-      else
-	 free(promote->data.string);
+      else {
+	 if (promote[pidx]->data.string != NULL) {
+	    free(promote[pidx]->data.string);
+	    promote[pidx]->data.string = NULL;
+	 }
+      }
 
       /* Promote numerical type to string */
       if (ops->type == XC_INT) {
-	 promote->data.string = (char *)malloc(13);
-	 sprintf(promote->data.string, "%12d", ops->parameter.ivalue);
-         nextptr = promote;
+	 promote[pidx]->data.string = (char *)malloc(13);
+	 sprintf(promote[pidx]->data.string, "%12d", ops->parameter.ivalue);
+         nextptr = promote[pidx++];
       }
       else if (ops->type == XC_FLOAT) {
-	 promote->data.string = (char *)malloc(13);
-	 sprintf(promote->data.string, "%g", (double)(ops->parameter.fvalue));
-         nextptr = promote;
+	 promote[pidx]->data.string = (char *)malloc(13);
+	 sprintf(promote[pidx]->data.string, "%g", (double)(ops->parameter.fvalue));
+         nextptr = promote[pidx++];
       }
       else {	/* ops->type == XC_EXPR */
 	 oparamptr ips;
 	 if (!compute_exprs && (ips = match_instance_param(localinst, key))
 		!= NULL && (ips->type == XC_STRING)) {
 	    nextptr = ips->parameter.string;
-	    /* This could probably be done better. . . hack to keep  */
-	    /* free() from operating twice on the same memory block. */
-	    promote->data.string = (char *)malloc(1);
+	    promote[pidx]->data.string = NULL;
 	 }
 	 else {
-	    promote->data.string = evaluate_expr(ops, localinst);
-            nextptr = promote;
+	    promote[pidx]->data.string = evaluate_expr(((localinst == NULL) ?
+			topobject : localinst->thisobject), ops, localinst);
+	    if (promote[pidx]->data.string != NULL)
+               nextptr = promote[pidx++];
+	    else
+	       nextptr = NULL;
 	 }
       }
+      pidx &= 0x1;	/* pidx toggles between 0 and 1 */
    }
    else
       nextptr = ops->parameter.string;
@@ -442,7 +458,8 @@ int findcurfont(int tpos, stringpart *strtop, objinstptr thisinst)
    stringpart *strptr;
 
    curpos = findstringpart(tpos, NULL, strtop, thisinst);
-   for (strptr = strtop; strptr != curpos; strptr = nextstringpart(strptr, thisinst))
+   for (strptr = strtop; (strptr != NULL) && (strptr != curpos);
+		strptr = nextstringpart(strptr, thisinst))
       if (strptr->type == FONT_NAME)
          cfont = strptr->data.font;
 
@@ -517,7 +534,7 @@ static char *nonprint[] = {
 	"Tab_Stop", "Tab_Forward", "Tab_Backward",
 	"Halfspace", "Quarterspace", "<Return>",
 	"Font", "Scale", "Color", "Kern", 
-        "Parameter", ">", "Error"};
+        "Parameter", ">", "Net_Name", "Error", NULL}; /* (jdk) */
 
 /*----------------------------------------------------------------------*/
 /* charprint():  							*/
@@ -545,7 +562,8 @@ void charprint(char *sout, stringpart *strptr, int locpos)
 	    *sout = '\0';
 	 break;
       case FONT_NAME:
-	 sprintf(sout, "Font=%s", fonts[strptr->data.font].psname);
+	 sprintf(sout, "Font=%s", (strptr->data.font >= fontcount) ?
+		"(unknown)" : fonts[strptr->data.font].psname);
 	 break;
       case FONT_SCALE:
 	 sprintf(sout, "Scale=%3.2f", strptr->data.scale);
@@ -693,8 +711,13 @@ char *textprintnet(char *prefix, char *pinstring, Genericlist *sublist)
 /*----------------------------------------------------------------------*/
 /* Test equivalence of the text string parts of a label with the	*/
 /* indicated char * string.						*/
-/*    If "exact" is True, requires an exact match, otherwise requires	*/
-/*    that the label only match text to the length of text.		*/
+/*									*/
+/* Return 0 if the strings match.  Return 1 or the result of strcmp()	*/
+/* on the first non-matching substring text segment if the strings	*/
+/* don't match.								*/
+/*									*/
+/* If "exact" is True, requires an exact match, otherwise requires	*/
+/* that the label only match text to the length of text.		*/
 /*----------------------------------------------------------------------*/
 
 int textcompx(stringpart *string, char *text, Boolean exact, objinstptr localinst)
@@ -796,6 +819,19 @@ Boolean issymbolfont(int fontnumber)
    return False;
 }
 
+/*----------------------------------------------------------------------*/
+/* Test if the specified font is in the "Helvetica" font family.	*/
+/* SVG uses this to make sure it scales down the font by 7/8 to match	*/
+/* our internal vectors, and to use "oblique" for the style instead of	*/
+/* "italic".								*/
+/*----------------------------------------------------------------------*/
+
+Boolean issansfont(int fontnumber)
+{
+   if (!strcmp(fonts[fontnumber].family, "Helvetica")) return True;
+   return False;
+}
+
 /*----------------------------------------------------------------------*/
 /* Test if the specified font is ISO-Latin1 encoding			*/
 /*----------------------------------------------------------------------*/
@@ -844,7 +880,7 @@ Boolean pin_is_bus(labelptr thislab, objinstptr thisinst)
 {
    stringpart *strptr;
    char *busptr;
-   int busidx;
+   /* int busidx; (jdk) */
    Boolean found_delimiter = FALSE;
 
    for (strptr = thislab->string; strptr != NULL; strptr =
@@ -1293,13 +1329,17 @@ stringpart *stringcopyback(stringpart *string, objinstptr thisinst)
 /* draw a single character at 0, 0 using current transformation matrix */
 /*---------------------------------------------------------------------*/
 
-short UDrawChar(u_char code, short styles, short ffont, int groupheight, int passcolor)
+short UDrawChar(u_char code, short styles, short ffont, int groupheight,
+	int passcolor, float passwidth)
 {
    objectptr drawchar;
    XPoint alphapts[2];
    short  localwidth;
    objinst charinst;
 
+   if ((ffont >= fontcount) || (fonts[ffont].encoding == NULL))
+      return 0;
+
    alphapts[0].x = 0;
    alphapts[0].y = 0;
    charinst.type = OBJINST;
@@ -1322,7 +1362,7 @@ short UDrawChar(u_char code, short styles, short ffont, int groupheight, int pas
 
    if (!(styles & 64)) {
       
-      UDrawObject(&charinst, SINGLE, passcolor, NULL);
+      UDrawObject(&charinst, SINGLE, passcolor, passwidth, NULL);
 
       /* under- and overlines */
       if (styles & 8)
@@ -1339,9 +1379,11 @@ short UDrawChar(u_char code, short styles, short ffont, int groupheight, int pas
 
 /*----------------------------------------------------------------------*/
 /* Draw an entire string, including parameter substitutions		*/
+/* (Normally called as UDrawString(); see below)			*/
 /*----------------------------------------------------------------------*/
 
-void UDrawString(labelptr drawlabel, int passcolor, objinstptr localinst)
+void UDrawString0(labelptr drawlabel, int passcolor, objinstptr localinst,
+	Boolean drawX)
 {
    stringpart *strptr;
    char *textptr;
@@ -1349,8 +1391,7 @@ void UDrawString(labelptr drawlabel, int passcolor, objinstptr localinst)
    int    pos, group = 0;
    int    defaultcolor, curcolor, scolor;
    short  oldx, oldfont, oldstyle;
-   float  tmpscale = 1.0, natscale = 1.0;
-   float  tmpthick = xobjs.pagelist[areawin->page]->wirewidth;
+   float  tmpscale = 1.0, natscale = 1.0, tmpthick = 2.0;
    XPoint newpoint, bboxin[2], bboxout[2];
    u_char xm, ym;
    TextExtents tmpext;
@@ -1387,7 +1428,7 @@ void UDrawString(labelptr drawlabel, int passcolor, objinstptr localinst)
    tmpjust = flipadjust(drawlabel->justify);
 
    /* "natural" (unscaled) length */
-   tmpext = ULength(drawlabel->string, localinst, 0.0, 0, NULL);
+   tmpext = ULength(drawlabel, localinst, 0, NULL);
 
    newpoint.x = (tmpjust & NOTLEFT ?
        (tmpjust & RIGHT ? -tmpext.width : -tmpext.width >> 1) : 0);
@@ -1420,7 +1461,7 @@ void UDrawString(labelptr drawlabel, int passcolor, objinstptr localinst)
 
        pos = 0;
        for (strptr = drawlabel->string; strptr != NULL;
-		strptr = nextstringpartnorecompute(strptr, localinst)) {
+		strptr = nextstringpart(strptr, localinst)) {
 
 	  /* All segments other than text cancel any	*/
 	  /* existing overline/underline in effect.	*/
@@ -1432,17 +1473,18 @@ void UDrawString(labelptr drawlabel, int passcolor, objinstptr localinst)
 
 	  switch(strptr->type) {
 	     case FONT_NAME:
-		ffont = strptr->data.font;
-		fstyle = 0;		   /* style reset by font change */
-	        if (baseline == newpoint.y) {  /* set top-level font and style */
-	           oldfont = ffont;
-	           oldstyle = fstyle;
-	        }
+		if (strptr->data.font < fontcount) {
+		   ffont = strptr->data.font;
+		   fstyle = 0;		   /* style reset by font change */
+	           if (baseline == newpoint.y) {  /* set top-level font and style */
+	              oldfont = ffont;
+	              oldstyle = fstyle;
+	           }
+		}
 		
 		/* simple boldface technique for derived fonts */
 
-		xobjs.pagelist[areawin->page]->wirewidth =
-   		   ((fonts[ffont].flags & 0x21) == 0x21) ?  4.0 : 2.0;
+		tmpthick = ((fonts[ffont].flags & 0x21) == 0x21) ?  4.0 : 2.0;
 		break;
 
 	     case FONT_SCALE:
@@ -1549,13 +1591,28 @@ void UDrawString(labelptr drawlabel, int passcolor, objinstptr localinst)
 		UPushCTM();
 		UPreMultCTM(DCTM, newpoint, tmpscale, 0);
 		addx = UDrawChar((u_char)32, fstyle, ffont, group,
-			curcolor);
+			curcolor, tmpthick);
 		newpoint.x += addx >> ((strptr->type == HALFSPACE) ? 1 : 2);
 		UPopCTM();
 		} break;
 
 	     case TEXT_STRING:
-		for (textptr = strptr->data.string; textptr && *textptr != '\0'; textptr++) {
+		textptr = strptr->data.string;
+
+		/* Don't write technology names in catalog mode if this	*/
+		/* option is enabled	(but *not* CATTEXT_MODE!)	*/
+
+		if (((eventmode == CATALOG_MODE) && !xobjs.showtech)
+			|| ((eventmode == CATTEXT_MODE)
+			&& (drawlabel != TOLABEL(EDITPART)))) {
+		   char *nsptr = strstr(textptr, "::");
+		   if (nsptr != NULL) {
+		      textptr = nsptr + 2;
+		      pos += (pointertype)nsptr - (pointertype)strptr->data.string + 2;
+		   }
+		}
+
+		for (; textptr && *textptr != '\0'; textptr++) {
 		   pos++;
 		   UPushCTM();
 		   UPreMultCTM(DCTM, newpoint, tmpscale, 0);
@@ -1565,14 +1622,14 @@ void UDrawString(labelptr drawlabel, int passcolor, objinstptr localinst)
 
 		   scolor = curcolor;
 	           if (passcolor == DOSUBSTRING) {
-	              if (pos <= textpos && pos > textend)
+	              if (pos <= areawin->textpos && pos > areawin->textend)
 		         scolor = SELECTCOLOR;
-		      else if (pos > textpos) {
+		      else if (pos > areawin->textpos) {
       		         XTopSetForeground(curcolor);
 		      }
 	           }
 	           newpoint.x += UDrawChar(*textptr, fstyle, ffont,
-			group, scolor) * tmpscale;
+			group, scolor, tmpthick) * tmpscale;
 
 		   UPopCTM();
 		}
@@ -1583,7 +1640,7 @@ void UDrawString(labelptr drawlabel, int passcolor, objinstptr localinst)
        }
    }
 
-enddraw:
+   /* enddraw: (jdk) */
 
    if (tabstops != NULL) free(tabstops);
 
@@ -1591,20 +1648,36 @@ enddraw:
 
    UPopCTM();
 
-   if (drawlabel->pin) UDrawXDown(drawlabel);
+   if (drawX && drawlabel->pin) UDrawXDown(drawlabel);
 
    if ((defaultcolor != DOFORALL) && (passcolor != curcolor)) {
       XTopSetForeground(passcolor);
    }
+}
 
-   xobjs.pagelist[areawin->page]->wirewidth = tmpthick;
+/*----------------------------------------------------------------------*/
+/* Draw String, with an "x" mark at the origin				*/
+/*----------------------------------------------------------------------*/
+
+void UDrawString(labelptr drawlabel, int passcolor, objinstptr localinst)
+{
+   UDrawString0(drawlabel, passcolor, localinst, TRUE);
+}
+
+/*----------------------------------------------------------------------*/
+/* Draw String, without the "x" mark					*/
+/*----------------------------------------------------------------------*/
+
+void UDrawStringNoX(labelptr drawlabel, int passcolor, objinstptr localinst)
+{
+   UDrawString0(drawlabel, passcolor, localinst, FALSE);
 }
 
 /*----------------------------------------------------------------------*/
 /* Compute the actual length of a string or portion thereof.		*/
 /*----------------------------------------------------------------------*/
 
-TextExtents ULength(stringpart *string, objinstptr localinst, float newscale,
+TextExtents ULength(labelptr drawlabel, objinstptr localinst,
 	short dostop, XPoint *tbreak)
 {
    float oldscale, strscale, natscale, locscale = 1.0, xtotal = 0.5;
@@ -1623,14 +1696,13 @@ TextExtents ULength(stringpart *string, objinstptr localinst, float newscale,
    if (fontcount == 0) return retext;
 
    /* Don't draw temporary labels from schematic capture system */
-   else if (string->type != FONT_NAME) return retext;
+   else if (drawlabel->string->type != FONT_NAME) return retext;
 
-   if (newscale > 0.0) natscale = UTopScale() * newscale;
-   else natscale = 1.0;
+   natscale = 1.0;
      
    oldscale = strscale = natscale;
 
-   for (strptr = string; strptr != NULL;
+   for (strptr = drawlabel->string; strptr != NULL;
 		strptr = nextstringpart(strptr, localinst)) {
       switch (strptr->type) {
 	 case SUPERSCRIPT:
@@ -1656,14 +1728,18 @@ TextExtents ULength(stringpart *string, objinstptr localinst, float newscale,
 	    xtotal = 0.5;
 	    break;
 	 case HALFSPACE: 
-	    chptr = (*(somebet + 32));
-	    xtotal += (float)(chptr->bbox.width + chptr->bbox.lowerleft.x)
+	    if (somebet) {
+	       chptr = (*(somebet + 32));
+	       xtotal += (float)(chptr->bbox.width + chptr->bbox.lowerleft.x)
 			* locscale * natscale / 2;
+	    }
 	    break;
 	 case QTRSPACE:
-	    chptr = (*(somebet + 32));
-	    xtotal += (float)(chptr->bbox.width + chptr->bbox.lowerleft.x) 
+	    if (somebet) {
+	       chptr = (*(somebet + 32));
+	       xtotal += (float)(chptr->bbox.width + chptr->bbox.lowerleft.x) 
 			* locscale * natscale / 4;
+	    }
 	    break;
 	 case TABBACKWARD:	/* find first tab value with x < xtotal */
 	    for (tabno = numtabs - 1; tabno >= 0; tabno--) {
@@ -1691,21 +1767,37 @@ TextExtents ULength(stringpart *string, objinstptr localinst, float newscale,
 	    strscale = natscale * strptr->data.scale;
 	    if (ykern == 0.0)
 	       natscale = strscale;
-   	    if (newscale > 0.0)
-	       strscale *= UTopScale();
 	    break;
 	 case KERN:
 	    xtotal += strptr->data.kern[0];
 	    ykern += strptr->data.kern[1];
 	    break;
 	 case FONT_NAME:
-	    somebet = fonts[strptr->data.font].encoding;
-	    locscale = fonts[strptr->data.font].scale;
-	    if (ykern == 0.0)
-	       natscale = locscale;
+	    if (strptr->data.font < fontcount) {
+	       somebet = fonts[strptr->data.font].encoding;
+	       locscale = fonts[strptr->data.font].scale;
+	       if (ykern == 0.0)
+	          natscale = locscale;
+	    }
 	    break;
 	 case TEXT_STRING:
-	    for (textptr = strptr->data.string; textptr && *textptr != '\0'; textptr++) {
+	    textptr = strptr->data.string;
+
+	    /* Don't write technology names in catalog mode if	*/
+	    /* the option is enabled, so ignore when measuring	*/
+
+	    if (((eventmode == CATALOG_MODE) && !xobjs.showtech)
+			|| ((eventmode == CATTEXT_MODE)
+			&& (drawlabel != TOLABEL(EDITPART)))) {
+		char *nsptr = strstr(textptr, "::");
+		if (nsptr != NULL) {
+		   textptr = nsptr + 2;
+		   locpos += (pointertype)nsptr - (pointertype)strptr->data.string + 2;
+		}
+	    }
+
+	    if (somebet == NULL) break;
+	    for (; textptr && *textptr != '\0'; textptr++) {
                if (dostop && (locpos >= dostop)) break;
 	       locpos++;
 
@@ -1732,7 +1824,7 @@ TextExtents ULength(stringpart *string, objinstptr localinst, float newscale,
 
    /* special case: return character position in retext.width */
    if (tbreak != NULL) {
-      int slen = stringlength(string, True, localinst);
+      int slen = stringlength(drawlabel->string, True, localinst);
       if ((tbreak->x - lasttotal) < (xtotal - tbreak->x))
 	 locpos = lastpos + 1;
       if (locpos < 1) locpos = 1;
@@ -1817,7 +1909,7 @@ void redrawtext(labelptr curlabel)
 
 void composefontlib(short cfont)
 {
-   genericptr *pgen;
+  /* genericptr *pgen; (jdk) */
    objinstptr *drawinst;
    objectptr *curlib, libobj, nullobj;
    objectptr directory = xobjs.libtop[FONTLIB]->thisobject;
@@ -1861,9 +1953,8 @@ void composefontlib(short cfont)
       instancedefaults(*drawinst, libobj,
 		(i % 16) * del + qdel,		/* X position */
 		-(i / 16) * del + qdel);	/* Y position */
-      drawinst = (objinstptr *)directory->plist + directory->parts;
+      drawinst = (objinstptr *)directory->plist + directory->parts - 1;
       (*drawinst)->color = DEFAULTCOLOR;
-      directory->parts++;
    }
 
    /* separate characters with gridlines (17 vert., 17 horiz.) */
@@ -1892,7 +1983,6 @@ void composefontlib(short cfont)
          pointptr->x = 16 * del;
          pointptr->y = (17 - i) * del;
       }
-      directory->parts++;
    }
 
    /* Set the bounding box for this display. 				*/
