@@ -67,7 +67,7 @@
 /*----------------------------------------------------------------------*/
 
 short	 popups;      /* total number of popup widgets on the screen */
-Boolean	 pressmode;   /* Whether we are in a press & hold state */
+int	 pressmode;   /* Whether we are in a press & hold state */
 
 xcWidget   top;
 xcWidget   message1, message2, message3, toolbar, overlay = NULL;
@@ -93,7 +93,6 @@ extern xcIntervalId printtime_id;
 extern short beeper;
 extern short fontcount;
 extern fontinfo *fonts;
-extern short textpos;
 extern short help_up;
 extern menustruct TopButtons[];
 #ifdef HAVE_XPM
@@ -156,6 +155,8 @@ static XtResource resources[] = {
 
   { "bboxcolor", "BBoxColor", XtRPixel, sizeof(Pixel),
       XtOffset(ApplicationDataPtr, bboxpix), XtRString, "greenyellow"},
+  { "clipcolor", "ClipColor", XtRPixel, sizeof(Pixel),
+      XtOffset(ApplicationDataPtr, clipcolor), XtRString, "powderblue"},
 
   /* GUI Color scheme 1 */
 
@@ -1019,7 +1020,10 @@ void outputpopup(xcWidget button, caddr_t clientdata, caddr_t calldata)
    sprintf(statics[5], "Orientation:");
    sprintf(statics[6], "Mode:");
 
-   sprintf(edit[0], "%s", curpage->filename);
+   if (curpage->filename)
+      sprintf(edit[0], "%s", curpage->filename);
+   else
+      sprintf(edit[0], "Page %d", areawin->page + 1);
    sprintf(edit[1], "%s", topobject->name);
 
    /* recompute bounding box and auto-scale, if set */
@@ -1302,7 +1306,7 @@ void outputpopup(xcWidget button, caddr_t clientdata, caddr_t calldata)
 /* For formatted strings, format first into _STR   */
 /*-------------------------------------------------*/
 
-xcTimeOutProc clrmessage(caddr_t clientdata, xcIntervalId *id)
+void clrmessage(XtPointer clientdata, xcIntervalId *id)
 {
    char buf1[50], buf2[50];
 
@@ -1367,7 +1371,7 @@ void xc_vprintf(xcWidget widget, const char *fmt, va_list args_in)
 
    /* 10 second timeout */
    if (widget == message3) {
-      printtime_id = xcAddTimeOut(app, 10000, (xcTimeOutProc)clrmessage, NULL);
+      printtime_id = xcAddTimeOut(app, 10000, clrmessage, NULL);
    }
 }
     
@@ -1974,6 +1978,14 @@ XCWindowData *GUI_init(int argc, char *argv[])
    newwin->gc = XCreateGC(dpy, newwin->window,
 		GCForeground | GCBackground | GCFont, &values);
 
+   newwin->clipmask = XCreatePixmap(dpy, newwin->window, newwin->width,
+		newwin->height, 1); 
+
+   values.foreground = 0;
+   values.background = 0;
+   newwin->cmgc = XCreateGC(dpy, newwin->clipmask, GCForeground
+		| GCBackground, &values);
+
    /* set the area widget width and height, center userspace (0,0) on screen */
 
    XtSetArg(wargs[0], XtNwidth, &newwin->width);
@@ -2024,7 +2036,7 @@ int main(int argc, char **argv)
 
    for (k = argc - 1; k > 0; k--) {
       if (!strncmp(argv[k], "-2", 2)) {
-	 pressmode = True;	/* 2-button mouse indicator */
+	 pressmode = 1;		/* 2-button mouse indicator */
 	 break;
       }
    }
@@ -2048,7 +2060,7 @@ int main(int argc, char **argv)
 #endif
 
    loadrcfile();
-   pressmode = False;	/* Done using this to mark 2-button mouse mode */
+   pressmode = 0;	/* Done using this to mark 2-button mouse mode */
 
    composelib(PAGELIB);	/* make sure we have a valid page list */
    composelib(LIBLIB);	/* and library directory */
